/* CRC.java
 *
 * Created on 29 September 2006, 09:37
 *
 * Alan Bradbury - AND Technology Research
 *
 * AXIS 64 programming utility.
 *
 */

package AXISProgrammer;

import java.util.*;

/** Stores and handles the CRC of a set of data.  The data is not stored in the  
 * CRC class, rather bytes are added to the CRC by class methods.
 * <p>
 * The CRC algorithm is:
 * <p>
 * Take each data byte in turn
 * <p>
 * a. XOR LSBit of data byte with LSBit of CRC - save reult for step c<p>
 * b. Rotate right the 16-bit CRC (the lowest bit is lost, the hightest bit set to zero)<p>
 * c. If result of (a) = 1 then XOR CRC with 0x8408
 * <p>
 * Rotate right the data byte and perform the above until all 8 bits have been processed
 *
 * @author Alan Bradbury - AND Technology Research
 * @version 1.0
 */
public class CRC {
    private static final int INITIAL_CRC = 0xFFFF;  // initial value of the CRC
    private static final int XOR_VALUE = 0x8408;    // XOR value defined in the algorithm
    private int CRCValue;   // the CRC value
    
    /** Creates a new instance of CRC */
    public CRC() {
        CRCValue = INITIAL_CRC;
    }

    /** Initialises the CRC */
    public void initialiseCRC() {
        CRCValue = INITIAL_CRC;
    }

    /** Adds a byte to the CRC 
     *
     * @param byteToAdd Byte to add to the CRC
     */
    public void addByteToCRC(int byteToAdd) {
        for (int nBit = 0; nBit < 8; nBit++) {  // loops through all bits of the byte
            boolean bXORLSBit = false;  // stores the result of alogrithm step (a) 
            
            // alogithm step (a)
            if (((CRCValue & 0x0001) ^ (byteToAdd & 0x01)) == 0x01) {
                bXORLSBit = true;
            }  
            
            CRCValue >>>= 1;    // alogrithm step (b)
            
            // algorithm step (c)
            if (bXORLSBit) {
                CRCValue ^= XOR_VALUE;
            }
            
            byteToAdd >>>= 1;   // rotate data byte right
        }
    }

    /** Gets the CRC value 
     *
     * @return The CRC value
     */
    public int getCRC() {
       return CRCValue; 
    }
}
