/* Programmer.java
 *
 * Created on 29 September 2006, 09:37
 *
 * Alan Bradbury - AND Technology Research
 *
 * AXIS 64 programming utility.
 *
 */

package AXISProgrammer;

import java.util.*;
import java.io.*;
import javax.swing.*;
import gnu.io.*;    // the serial comms functions

import java.awt.event.*;

/** Handles programming a processor.
 *
 * @author Alan Bradbury - AND TechnologyResearch
 * @version 1.0
 */
public final class Programmer implements SerialPortEventListener, ActionListener {
	// constants
	private static final int BAUDRATE = 57600;							// the comm port baud rate
	private static final int DATABITS = SerialPort.DATABITS_7;			// the number of data bits for the comm port
	private static final int STOPBITS = SerialPort.STOPBITS_1;			// the number of stop bits for the comm port
	private static final int PARITY = SerialPort.PARITY_EVEN;			// the parity for the comm port
	private static final int FLOWCONTROL = SerialPort.FLOWCONTROL_NONE;	// the flow control for the comm port
	
    /** Key processor 1 programming packet ID */
    public static final int KEY_PROCESSOR_1 = 0x00;

    /** Key processor 2 programming packet ID */
    public static final int KEY_PROCESSOR_2 = 0x10;

    /** Key processor 3 programming packet ID */
    public static final int KEY_PROCESSOR_3 = 0x20;

    /** Function key processor programming packet ID */
    public static final int FUNCTION_KEY_PROCESSOR = 0x30;

    /** Main processor programming packet ID */
    public static final int MAIN_PROCESSOR = 0x40;

    /** Velocity table programming packet ID */
    public static final int VELOCITY_TABLE = 0x50;
    
    private static final int SOH = 0x01;    // program packet header
    private static final int ENTER_PROGRAM_MODE = (MAIN_PROCESSOR | 0x0B); // enter program mode command
    private static final int EXIT_PROGRAM_MODE = (MAIN_PROCESSOR | 0x0A); // exit program mode command
    private static final int ENQUIRE_PROGRAM_MODE = (MAIN_PROCESSOR | 0x08); // enquire program mode status command
    private static final int IN_PROGRAM_MODE = (MAIN_PROCESSOR | 0x0F); // in program mode reply
    private static final int NOT_IN_PROGRAM_MODE = (MAIN_PROCESSOR | 0x0E); // not in program mode reply
    private static final int LINE_ACK = 0x0C;   // ACK after downloading a line
    private static final int LINE_NAK = 0x0D;   // NAK after downloading a line
    private static final int GET_CRC = 0x09;    // get CRC command
    private static final int CONTAINS_CRC = 0x04;    // indicates the packet contains a CRC
    
    private static final int DELAY_OPENING_COMM_PORT = 2000;    // the delay (in ms) to open the comm port (2s)
    private static final int COMM_PORT_REPLY_WAIT = 2000;    // time to wait (in ms) for a reply (2s)
    private static final int DOWNLOAD_LINE_RETRIES = 5; // number of times to retry downloading a line
    
    private static final int MAXIMUM_PROGRAMMING_ADDRESS = 0x7000;  // the maximum programming address above which data should not be sent
    
    private static final int NUM_HEX_FILE_HEADER_CHAR = 9;  // the number of characters in a hex file line header

	private ProgressBarUpdater progressBarUpdater;	// indicates to the user the progress of programming
	private SerialPort serialPort = null;   // the serial port to program with
	private String portName;
	private int totalNumberOfLinesDownloaded = 0;
    
    private CRC programDataCRC = new CRC(); // CRC of the data being programmed
	
	private TextAreaUpdater output = null;
	
	private Object lock = new Object();
	private int eventCount = 0;
	private javax.swing.Timer timer = null;
    
    /** Creates a new instance of Programmer 
     *
     * @param userProperties The programs user properties (including comm port properties)
     * @param textArea Text area to output error messages to
     */
	protected Programmer(String portName, TextAreaUpdater output, ProgressBarUpdater progressBarUpdater) {
		this.portName = portName;
		this.output = output;
		this.progressBarUpdater = progressBarUpdater;
		totalNumberOfLinesDownloaded = 0;
	}

    /** Downloads a line to program */
    private boolean downloadLine(String lineToDownload, int programID) {
        try {
            char packetChecksum = 0x00; // stores the checksum of the packet
            int lineIndex;  // index into the line characters
            int numberOfDataCharacters = Integer.parseInt(lineToDownload.substring(1,3), 16) * 2;
            //int returnedValue;
            //int returnedByte;
            //long replyTime = 0;
            //Clock replyTimer = new Clock();

            serialPort.getOutputStream().write(SOH);    // packet header
            packetChecksum += SOH;
            serialPort.getOutputStream().write(numberOfDataCharacters + NUM_HEX_FILE_HEADER_CHAR + 6);   // number of packet bytes
            packetChecksum += (numberOfDataCharacters + NUM_HEX_FILE_HEADER_CHAR + 6);
            serialPort.getOutputStream().write(programID);   // packet command
            packetChecksum += programID;

            // download the line header

            for (lineIndex = 0; lineIndex < NUM_HEX_FILE_HEADER_CHAR; lineIndex++) {    // itterates through the lines header characters
                serialPort.getOutputStream().write(lineToDownload.charAt(lineIndex));   // sends the line header characters
                packetChecksum += lineToDownload.charAt(lineIndex);
            }

            // download the line data

            for (lineIndex = NUM_HEX_FILE_HEADER_CHAR; lineIndex < NUM_HEX_FILE_HEADER_CHAR + numberOfDataCharacters; lineIndex++) {    // itterates through the lines data characters
                serialPort.getOutputStream().write(lineToDownload.charAt(lineIndex));   // sends the line data characters
                packetChecksum += lineToDownload.charAt(lineIndex);
            }

            // add the line data to the CRC - the line data is added as the bytes to program not as the characters sent
            // i.e if the first two data bytes are 20 then we send the character 2 followed by 0 but the value added to
            // the CRC is 0x2 = 32

            for (lineIndex = NUM_HEX_FILE_HEADER_CHAR; lineIndex < NUM_HEX_FILE_HEADER_CHAR + numberOfDataCharacters; lineIndex+=2) {    // itterates through the lines data characters
                programDataCRC.addByteToCRC(Integer.parseInt(lineToDownload.substring(lineIndex,lineIndex+2), 16));
            }

            // download the line checksum

            serialPort.getOutputStream().write(lineToDownload.charAt(NUM_HEX_FILE_HEADER_CHAR + numberOfDataCharacters));   // sends the line checksum characters
            packetChecksum += lineToDownload.charAt(NUM_HEX_FILE_HEADER_CHAR + numberOfDataCharacters);
            serialPort.getOutputStream().write(lineToDownload.charAt(NUM_HEX_FILE_HEADER_CHAR + numberOfDataCharacters + 1));   // sends the line checksum characters
            packetChecksum += lineToDownload.charAt(NUM_HEX_FILE_HEADER_CHAR + numberOfDataCharacters + 1);

            // send the packet checksum

            serialPort.getOutputStream().write(packetChecksum & 0x7F);
            
            // get the line reply packet
            
			if (waitAvailable(4, COMM_PORT_REPLY_WAIT)) {
            //replyTimer.reset();
            //while (((serialPort.getInputStream().available()) != 4) && ((replyTime = replyTimer.getTime()) < COMM_PORT_REPLY_WAIT)) {}    // waits for the return value 
            
            //if (replyTime < COMM_PORT_REPLY_WAIT) {
                int returnedByte = serialPort.getInputStream().read();
                if (returnedByte == SOH) {
                    returnedByte = serialPort.getInputStream().read();
                    if (returnedByte == 4) {
                        int returnedValue = serialPort.getInputStream().read();

                        returnedByte = serialPort.getInputStream().read();
                        if (returnedByte == ((returnedValue + 4 + SOH) & 0x7F)) { // is the packet checksum correct
                            if (returnedValue == (LINE_ACK | programID)) {    // has a line successfully programmed
                                // a line has successfully programmed

                                return true;
                            }
                            else if (returnedValue == (LINE_NAK | programID)) {
                                // a line has failed to program

                                writeOutput("> Error: NAK received\r\n");
                            }
                        }
                        else {
                            writeOutput("> Error: Packet checksum incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                        }
                    }
                    else {
                        writeOutput("> Error: Packet data length incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                    }
                }
                else {
                    writeOutput("> Error: Packet header byte incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                }
            }
            else {
                int charactersAvailable = serialPort.getInputStream().available();
                String outputString = Integer.toString(charactersAvailable);

                while (charactersAvailable != 0) {
                    outputString += (", 0x" + Integer.toString(serialPort.getInputStream().read(), 16));
                    charactersAvailable--;
                }

                writeOutput("> Error: Waiting for a reply timed out (" + outputString + ")\r\n");
            }
        }
		catch (IOException ex) {
            writeOutput("> Error: Serial port I/O exception\r\n");
        }
        
        return false;
    }
	
	private final void setProgressBarValue(int value) {
		progressBarUpdater.setValue(value);
	}

    /** checks that the lines address is in range */
    private boolean checkLineAddress(String lineToCheck) {
        if (lineToCheck != null) {
            int lineAddress = Integer.parseInt(lineToCheck.substring(3, 7), 16);

            if (lineAddress < MAXIMUM_PROGRAMMING_ADDRESS) {
                return true;
            }
        }
        
        return false;
    }
	
	private boolean timerRunning;
	
	// modified to fix the bug...
	/*
	private final boolean waitAvailable(int count, int ms) {
		try {
			InputStream inputStream = serialPort.getInputStream();
			
			int available = inputStream.available();
			if (available >= count) {
				return true;
			}
			else {
				timerRunning = true;
				
				// start the timer
				java.util.Timer t = new java.util.Timer();
				t.schedule(new java.util.TimerTask() {
					public final void run() {
						synchronized (lock) {
							timerRunning = false;
							lock.notify();
						}
					}}, ms);
				
				boolean success = false;
				while (true) {
					available = inputStream.available();
					if (available >= count) {
						success = true;
						break;		// escape from loop
					}
					else {
						synchronized (lock) {
							if (timerRunning) {
								try {
									// wait 10ms, or until timeout occurs before re-sampling characters
									lock.wait(10);
								}
								catch (InterruptedException e) {
									break;		// escape from loop
								}
							}
							else {
								break;		// escape from loop
							}
						}
					}
				}
				
				t.cancel();			// cancel the timer
				return success;
			}
		}
		catch (IOException e) {
			return false;
		}
	}
	*/
	
	private final boolean waitAvailable(int count, int ms) {
		try {
			InputStream inputStream = serialPort.getInputStream();
			
			int available = inputStream.available();
			if (available >= count) {
				return true;
			}
			else {
				// start the timer
				synchronized (lock) {
					timer = new javax.swing.Timer(ms, this);
					timer.setRepeats(false);
					timer.start();
				}
				
				while (true) {
					synchronized (lock) {
						eventCount = 0;
					}
					available = inputStream.available();
					if (available >= count) {
						timer.stop();
						return true;
					}
					else if (!timer.isRunning()) {
						return false;
					}
					else {
						synchronized (lock) {
							if (eventCount == 0) {
								try {
									lock.wait();
								}
								catch (InterruptedException e) {
									timer.stop();
									return false;
								}
							}
						}
					}
				}
			}
		}
		catch (IOException e) {
			return false;
		}
	}
	
	public final void actionPerformed(ActionEvent e) {
		// called from the event dispatcher thread
		synchronized (lock) {
			Object source = e.getSource();
			if (source == timer) {
				lock.notify();
			}
		}
	}
    
    /** checks gets and checks that the CRC is correct */
    private boolean checkCRC(int programID) {
        //int returnedValue;
        //int returnedByte;
        //long replyTime = 0;
        //Clock replyTimer = new Clock();

        // send the get CRC command

        try {
            serialPort.getOutputStream().write(SOH);    // header
            serialPort.getOutputStream().write(0x04);   // number of bytes
            serialPort.getOutputStream().write(GET_CRC | programID);   // command
            serialPort.getOutputStream().write(SOH + 4 + (GET_CRC | programID));   // checksum

			if (waitAvailable(8, COMM_PORT_REPLY_WAIT)) {
            //replyTimer.reset();
            //while (((serialPort.getInputStream().available()) != 8) && ((replyTime = replyTimer.getTime()) < COMM_PORT_REPLY_WAIT)) {}    // waits for the return value 

            //if (replyTime < COMM_PORT_REPLY_WAIT) {
                int returnedByte = serialPort.getInputStream().read();
                if (returnedByte == SOH) {
                    returnedByte = serialPort.getInputStream().read();
                    if (returnedByte == 8) {
                        returnedByte = serialPort.getInputStream().read();
                        if (returnedByte == (CONTAINS_CRC | programID)) {
                            byte[] returnedDataBytes = new byte[4];

                            serialPort.getInputStream().read(returnedDataBytes);
                            int returnedValue = Integer.parseInt(new String(returnedDataBytes), 16);

                            returnedByte = serialPort.getInputStream().read();
                            if (returnedByte == ((returnedDataBytes[0] + returnedDataBytes[1] + returnedDataBytes[2] + returnedDataBytes[3] + 8 + SOH + (CONTAINS_CRC | programID)) & 0x7F)) { // is the packet checksum correct
                                if (returnedValue == programDataCRC.getCRC()) {    // is the CRC correct
                                    // the CRC is correct

                                    return true;
                                }
                                else {
                                    writeOutput("> Error: CRC not correct (0x" + Integer.toString(returnedValue, 16) + ", 0x" + Integer.toString(programDataCRC.getCRC(), 16) + ")\r\n");
                                }
                            }
                            else {
                                writeOutput("> Error: Packet checksum incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                            }
                        }
                        else {
                            writeOutput("> Error: Packet ID incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                        }
                    }
                    else {
                        writeOutput("> Error: Packet data length incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                    }
                }
                else {
                    writeOutput("> Error: Packet header byte incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                }
            }
            else {
                int charactersAvailable = serialPort.getInputStream().available();
                String outputString = Integer.toString(charactersAvailable);

                while (charactersAvailable != 0) {
                    outputString += (", 0x" + Integer.toString(serialPort.getInputStream().read(), 16));
                    charactersAvailable--;
                }

                writeOutput("> Error: Waiting for a reply timed out (" + outputString + ")\r\n");
            }
        }
		catch (IOException ex) {
            writeOutput("> Error: Serial port I/O exception\r\n");
        }
        return false;
    }

    /** Counts the number of lines in a download file in the correct address range
     *
     * @param fileToCount File whose lines are to be counted
     * @return Number of lines in the file
     */
    public int countLinesInFile(BufferedReader fileToCount) {
        int numberOfLines = 0;
        
        try {
            fileToCount.readLine();    // skip the first line

            String nextLine;
            while (checkLineAddress(nextLine = fileToCount.readLine())) {   // itterates through the lines in the file in the correct address range
                numberOfLines++;
            }
            
            return numberOfLines;
        }
		catch (IOException ex) {
            writeOutput("> Error: Reading download file line\r\n");
            return 0;
        }
    }
    
    /** Downloads a file to program
     *
     * @param downloadFileBuffer The file to progarm
     * @param programID The porgramming packet ID of the area to program
     * @param totalNumberOfLines The total number of lines to download for all files
     * @return True if successful, false if not
     */
    public boolean downloadFile(BufferedReader downloadFileBuffer, int programID, int totalNumberOfLines) {
        try {
            char[] lineCharacters = new char[21];
            int downloadLineRetries;    // the number of times to retring a line
            int numberLinesDownload = 0;    // the number of lines downloaded
            
            programDataCRC.initialiseCRC();

            downloadFileBuffer.readLine();    // skip the first line
            
            String nextLine;
            while (checkLineAddress(nextLine = downloadFileBuffer.readLine())) {   // itterates through the lines in the update file in the correct address range
                downloadLineRetries = DOWNLOAD_LINE_RETRIES;
                while (downloadLine(nextLine, programID) == false) {   // did the line fail to download
                    // the line failed to download

                    while (serialPort.getInputStream().read() != -1) {} // clears the serial port input stream

                    downloadLineRetries--;    
                    if (downloadLineRetries == 0) {
                        writeOutput("> Error: Failed downloading a line (" + nextLine + ", " + Integer.toString(numberLinesDownload) + ")\r\n");
                        return false;
                    }
                }
                numberLinesDownload++;  // the line was downloaded successfully so increment the number of lines downloaded
                totalNumberOfLinesDownloaded++;
                setProgressBarValue((int)((totalNumberOfLinesDownloaded * 100) / totalNumberOfLines));
            }

            // check the CRC

            writeOutput("> Checking CRC (" + Integer.toString(numberLinesDownload) + ")...\r\n");
            if (checkCRC(programID)) {
                // we have programmed the file successfully
                
                writeOutput("> Downloaded file successfully (0x" + Integer.toString(programID, 16) + ")\r\n");
                return true;
            }
        }
		catch (IOException ex) {
            writeOutput("> Error: Serial port I/O exception\r\n");
        }
        
        writeOutput("> Error: Failure downloading a file (0x" + Integer.toString(programID, 16) + ")\r\n");
        return false;
    }

    /** Connects/disconnects the programmer to the programming comm port 
     *
     * @param connect Indicates whether to connect or disconnect to the programming comm port
     * @return True if successful, false if not
     */
    public boolean connect(boolean connect) {
        if (connect) {  // are we connecting to the programming comm port
            // connect to the programming comm port

			// get the comm port
			CommPortIdentifier portIdentifier = null;
		//	String portName = userProperties.getPortName();
			if (portName != null) {
				try {
					portIdentifier = CommPortIdentifier.getPortIdentifier(portName);
				}
				catch (NoSuchPortException e) {
					portIdentifier = null;
				}
			}
			
            if (portIdentifier != null) {  // is a comm port selected
                try {
                    CommPort commPort = portIdentifier.open(UpdateDialog.class.getName(), DELAY_OPENING_COMM_PORT);  // open the port and wait 2s for this to succeed
                    
                    if (commPort instanceof SerialPort) {   // check that the comm port is a serial port
                        serialPort = (SerialPort)commPort;
 
                        // add the serial port event listener
                        
                        try {
                            serialPort.addEventListener(this);
                            serialPort.notifyOnDataAvailable(true); // enable data available event
                            serialPort.notifyOnParityError(true); // enable parity error event

                            // set the serial port properties

                            try {
                                serialPort.setSerialPortParams(BAUDRATE, DATABITS, STOPBITS, PARITY);
                                serialPort.setFlowControlMode(FLOWCONTROL);
                                writeOutput("> " + portIdentifier.getName() + " opened\r\n");
                                return true;
                            }
							catch (UnsupportedCommOperationException err) {
                                writeOutput("> Error: Unable to set " + portIdentifier.getName() + "s properties\r\n");
                                commPort.close();
                                serialPort = null;
                            }
                        }
                        catch (TooManyListenersException err) {
                            writeOutput("> Error: Unable to add event lisitner to " + portIdentifier.getName() + "\r\n");
                            commPort.close();
                            serialPort = null;
                        }
                    }
					else {
                        writeOutput("> Error: " + portIdentifier.getName() + " is not a serial port\r\n");
                        commPort.close();
                    }
                }
				catch (PortInUseException err) {
                    writeOutput("> Error: " + portIdentifier.getName() + " is currently open (" + err.currentOwner + ")\r\n");
                }
            }
            else {
                writeOutput("> Error: No comm port selected\r\n");
            }
        }
		else {
            // disconnect from the programming comm port
            
            if (serialPort != null) {   // is a serial port open
                serialPort.close();
			//	String portName = userProperties.getPortName();
				String name = portName;
				if (name == null) {
					name = "";
				}
                writeOutput("> " + name + " closed\r\n");
                return true;
            }
        }

        return false;
    }
    
    /** Responds to the serial port events.  Currently responds to the following events:
     * <p>
     * parity error
     * 
     * @param event The serial port event
     */
    public void serialEvent(SerialPortEvent event) {
		// called from the com port thread!!!
		synchronized (lock) {
			lock.notify();		// just do this!!!
		}
		
		int type = event.getEventType();
		if (type == event.PE) {   // parity error
			writeOutput("> Error: Parity\r\n");
		}
    }
    
    /** Enters program mode 
     *
     * @return True if successful, false if not
     */
    public boolean enterProgramMode() {
        try {
            writeOutput("> Entering program mode\r\n");

            // send the enter programming mode command
            
            serialPort.getOutputStream().write(SOH);    // header
            serialPort.getOutputStream().write(0x04);   // number of bytes
            serialPort.getOutputStream().write(ENTER_PROGRAM_MODE);   // command
            serialPort.getOutputStream().write(0x50);   // checksum

            // wait one second to enter programming mode

            try {
                Thread.sleep(1000);

                if (getProgramModeStatus() == true) {    // are we in programming mode
                    // we are in programming mode
                
                    return true;
                }
            }
			catch (InterruptedException ex) {
                writeOutput("> Error: Sleep interrupted exception\r\n");
            }
        }
		catch (IOException ex) {
            writeOutput("> Error: Serial port I/O exception\r\n");
        }
        
        writeOutput("> Error: Failure entering programming mode\r\n");
        return false;
    }
          
    /** Exits program mode */
    public void exitProgramMode() {
        try {
            // send the exit programming mode command
            
            serialPort.getOutputStream().write(SOH);    // header
            serialPort.getOutputStream().write(0x04);   // number of bytes
            serialPort.getOutputStream().write(EXIT_PROGRAM_MODE);  // command
            serialPort.getOutputStream().write(0x4F);   // checksum
            writeOutput("> Exited program mode\r\n");
        }
		catch (IOException ex) {
            writeOutput("> Error: Serial port I/O exception\r\n");
            writeOutput("> Error: Failure exiting programming mode\r\n");
        }
    }

    /** Get program mode status
     *
     * @return True if in program mode, false if not
     */
    public boolean getProgramModeStatus() {
        try {
            //int returnedByte;
            //int returnedValue;
            //long replyTime = 0;
            //Clock replyTimer = new Clock();
            
            // send the get programming mode command

            serialPort.getOutputStream().write(SOH);    // header
            serialPort.getOutputStream().write(0x04);   // number of bytes
            serialPort.getOutputStream().write(ENQUIRE_PROGRAM_MODE);   // command
            serialPort.getOutputStream().write(0x4D);   // checksum
            
			if (waitAvailable(4, COMM_PORT_REPLY_WAIT)) {
            //replyTimer.reset();
            //while (((serialPort.getInputStream().available()) != 4) && ((replyTime = replyTimer.getTime()) < COMM_PORT_REPLY_WAIT)) {}    // waits for the return value 
            
            //if (replyTime < COMM_PORT_REPLY_WAIT) {
                int returnedByte = serialPort.getInputStream().read();
                if (returnedByte == SOH) {
                    returnedByte = serialPort.getInputStream().read();
                    if (returnedByte == 4) {
                        int returnedValue = serialPort.getInputStream().read();

                        returnedByte = serialPort.getInputStream().read();
                        if (returnedByte == ((returnedValue + 4 + SOH) & 0x7F)) { // is the packet checksum correct
                            if (returnedValue == IN_PROGRAM_MODE) {    // are we in programming mode
                                // we are in programming mode

                                writeOutput("> In programming mode\r\n");
                                return true;
                            }
                            else if (returnedValue == NOT_IN_PROGRAM_MODE) {
                                // we are not in programming mode

                                writeOutput("> Not in programming mode\r\n");
                                return false;
                            }
                            else {
                                writeOutput("> Error: Packet data incorrect (0x" + Integer.toString(returnedValue, 16) + ")\r\n");
                            }
                        }
                        else {
                            writeOutput("> Error: Packet checksum incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                        }
                    }
                    else {
                        writeOutput("> Error: Packet data length incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                    }
                }
                else {
                    writeOutput("> Error: Packet header byte incorrect (0x" + Integer.toString(returnedByte, 16) + ")\r\n");
                }
            }
            else {
                int charactersAvailable = serialPort.getInputStream().available();
                String outputString = Integer.toString(charactersAvailable);

                while (charactersAvailable != 0) {
                    outputString += (", 0x" + Integer.toString(serialPort.getInputStream().read(), 16));
                    charactersAvailable--;
                }
                
                writeOutput("> Error: Waiting for a reply timed out (" + outputString + ")\r\n");
            }
        }
		catch (IOException ex) {
            writeOutput("> Error: Serial port I/O exception\r\n");
        }

        writeOutput("> Error: Failure getting programming mode status\r\n");
        return false;
    }
	
	private final void writeOutput(String text) {
		output.append(text);
	}
}
